# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""The Sloecode code hosting file system.

TODO: write something interesting.

"""

__metaclass__ = type
__all__ = [
    'get_BZR_PLUGIN_PATH_for_subprocess',
    'get_sc_server',
    ]

import os
import xmlrpclib

from bzrlib.errors import PermissionDenied
from bzrlib.transport import (
    register_transport,
    Server,
    unregister_transport,
    )

from sloecode.codehosting.transport import (
    VirtualTransport,
    get_chrooted_transport,
    get_readonly_transport,
    )


def get_bzr_plugins_path():
    """Find the path to the Bazaar plugins for this rocketfuel instance."""
    here_location = os.path.abspath(__file__)
    codehosting_dir = os.path.dirname(here_location)
    sloecode_dir = os.path.dirname(codehosting_dir)
    return os.path.join(sloecode_dir, 'bzrplugins')


def get_BZR_PLUGIN_PATH_for_subprocess():
    """Calculate the appropriate value for the BZR_PLUGIN_PATH environment.

    The '-site' token tells bzrlib not to include the 'site specific plugins
    directory' (which is usually something like
    /usr/lib/pythonX.Y/dist-packages/bzrlib/plugins/) in the plugin search
    path, which would be inappropriate for Launchpad, which may be using a bzr
    egg of an incompatible version.
    """
    return ":".join((get_bzr_plugins_path(), "-site"))


def get_path_segments(path, maximum_segments=-1):
    """Break up the given path into segments.

    If 'path' ends with a trailing slash, then the final empty segment is
    ignored.
    """
    return path.strip('/').split('/', maximum_segments)


class SloecodeServer(Server):
    """Bazaar `Server` for translating Sloecode paths."""

    def __init__(self, codehosting_client, username,
                 user_transport, project_transport,
                 seen_new_branch_hook=None):
        """Construct a SloecodeServer.

        :param scheme: The URL scheme to use.
        :param codehosting_api: An XML-RPC client that implements
            getAccessPermissionsForUser.
        :param username: The authenticated username for the user who is
            accessing branches.
        :param user_transport: A chrooted transport to the directory
            containing the user repositories.
        :param project_directory: A chrooted transport to the directory
            contaning the project repositories.
        :param seen_new_branch_hook: A callable that will be called once for
            each branch accessed via this server.
        """
        self.scheme = 'sc-%d:///' % id(self)
        self.codehosting_client = codehosting_client
        self.username = username
        self.user_transport = user_transport
        self.project_transport = project_transport
        self._is_start_server = False
        self.is_admin = False
        self.readable_projects = []
        self.writable_projects = []
        self._is_started = False

    def _transportFactory(self, url):
        """Create a transport for this server pointing at `url`.

        This constructs a regular Bazaar `Transport` from the "asynchronous
        transport" factory specified in the `asyncTransportFactory` instance
        variable.
        """
        assert url.startswith(self.scheme)
        return VirtualTransport(self, url)

    def _queryUserPermissions(self):
        self.is_admin, self.readable_projects, self.writable_projects = (
            self.codehosting_client.getAccessPermissionsForUser(self.username))

    def start_server(self):
        """Start the server and query the permissions for the user."""
        if self._is_started:
            return
        register_transport(self.scheme, self._transportFactory)
        self._is_started = True
        self._queryUserPermissions()

    def stop_server(self):
        """See Server.stop_server."""
        if not self._is_started:
            return
        self._is_started = False
        unregister_transport(self.scheme, self._transportFactory)

    def translateUserPath(self, username):
        if username != self.username and not self.is_admin:
            raise PermissionDenied(
                "You (%s) do not have access to user %s's branches." %
                self.username, username)
        transport = self.user_transport.clone(username)
        return transport

    def translateProjectPath(self, project_name):
        if project_name in self.writable_projects:
            transport = self.project_transport
        elif project_name in self.readable_projects:
            transport = get_readonly_transport(self.project_transport)
        else:
            raise PermissionDenied(
                "You do not have access to branches for %s." % project_name)
        return transport.clone(project_name)

    def translateVirtualPath(self, virtual_url_fragment):
        """See `AsyncVirtualServer.translateVirtualPath`.

        Call 'translatePath' on the branchfs client with the fragment and then
        use 'makeTransport' on the _transport_dispatch to translate that
        result into a transport and trailing path.
        """
        segments = get_path_segments(virtual_url_fragment, 1)
        if len(segments) == 1:
            first = segments[0]
            trailing = None
        elif len(segments) == 2:
            first, trailing = segments
        else:
            raise RuntimeError('unexpected segments: %r' % segments)
        if first.startswith('~'):
            transport = self.translateUserPath(first[1:])
        else:
            transport = self.translateProjectPath(first)
        return transport, trailing


def get_sc_server(user_id,
                  user_directory,
                  project_directory,
                  server_endpoint,
                  seen_new_branch_hook=None):
    """Create a Sloecode server.

    :param user_id: The authenticated username.
    :param user_directory: The directory where the user repositories are
        located.
    :param project_directory: The directory where the project repositories
        are located.
    :param server_endpoint: URL for the branch file system end-point.
    :param seen_new_branch_hook: Normally used for changing the process title.
    :return: A `SloecodeServer`.
    """
    user_transport = get_chrooted_transport(user_directory)
    project_transport = get_chrooted_transport(project_directory)
    codehosting_client = xmlrpclib.ServerProxy(server_endpoint)

    sc_server = SloecodeServer(
        codehosting_client,
        user_id,
        user_transport,
        project_transport,
        seen_new_branch_hook)
    return sc_server
