"""XMLRPC services for Sloecode.
"""

import logging

from pylons.controllers.xmlrpc import XMLRPCController
from sloecode.model import Person, Project
from sloecode.model.site_role import PROJECT_ADMIN
from sloecode.model.membership import PR_DEVELOPER, PR_MANAGER


log = logging.getLogger(__name__)


class XmlrpcController(XMLRPCController):
    """Provides an XMLRPC interface to the sloecode database. This is the main
    entry point for other processes to get authentication and authorisation data
    from the sloecode database.
    """

    def getKeysForUser(self, username):
        """Get all ssh keys for a given username. The return value is a list of
        keys, each key being represented by a list of parts:
            [type, data, comment]

        If the requested username is not present, and empty list will be
        returned.
        """
        log.info("getKeysForUser: %r" % (username, ))
        ppl = Person.get(login=username)
        user_dict = {}
        keys = []
        if len(ppl) == 1:
            for key in ppl[0].auth_keys:
                keys.append([key.key_type, key.key_data, key.key_comment])
            user_dict[username] = keys

        log.info("Returning user_dict: %r" % (user_dict, ))
        return user_dict

    getKeysForUser.signature = [['struct', 'string']]

    def getAccessPermissionsForUser(self, username):
        """Get access permissions for a Sloecode user.

        This function returns a tuple that contains three parts:
            1. A boolean, specifying whether the user is a project admin.
            2. A list of projects the user has read-only access to.
            3. A list of projects the user has write-only access to.

        Note that if the user is a project admin, the second part of the list
        will be empty (since project admins can write to EVERYTHING), and the
        last part of the tuple will contain every project on the system.
        """
        log.info("getAccessPermissionsForUser: %r" % (username, ))
        # get person object:
        ppl = Person.get(login=username)
        # if we got a bad username, just return empty lists:
        if not ppl:
            return [False, [], []]

        # work out if the person is a project admin.
        isAdmin = False
        person = ppl[0]
        for sr in person.site_roles:
            if sr.role == PROJECT_ADMIN:
                isAdmin = True
                break

        readable_projects = []
        writable_projects = []
        # If user is a project admin, we return all project names under the
        # writable section (since they are admins they can write to everything).
        if isAdmin:
            writable_projects = [p.name for p in Project.get()]
        else:
            for mem in person.projects:
                if mem.role in (PR_DEVELOPER, PR_MANAGER):
                    writable_projects.append(mem.project.name)
                else:
                    readable_projects.append(mem.project.name)
        return [isAdmin, readable_projects, writable_projects]

    getAccessPermissionsForUser.signature = [['array', 'string']]
