# Copyright 2009 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""SSH session implementations for the codehosting SSH server."""

__metaclass__ = type
__all__ = [
    'SmartServerLauncher',
    ]

import os

from zope.event import notify

from twisted.internet import process
from twisted.python import log

from sloecode.codehosting.server import get_BZR_PLUGIN_PATH_for_subprocess
from sloecode.sshserver.events import AvatarEvent
from sloecode.sshserver.session import DoNothingSession


class BazaarSSHStarted(AvatarEvent):

    template = '[%(session_id)s] %(username)s started bzr+ssh session.'


class BazaarSSHClosed(AvatarEvent):

    template = '[%(session_id)s] %(username)s closed bzr+ssh session.'


class ForbiddenCommand(Exception):
    """Raised when a session is asked to execute a forbidden command."""


class _WaitForExit(process.ProcessReader):
    """Wait on a socket for the exit status."""

    def __init__(self, reactor, proc, sock):
        super(_WaitForExit, self).__init__(reactor, proc, 'exit',
                                           sock.fileno())
        self._sock = sock
        self.connected = 1

    def close(self):
        self._sock.close()

    def dataReceived(self, data):
        # TODO: how do we handle getting only *some* of the content?, Maybe we
        #       need to read more bytes first...

        # This is the only thing we do differently from the standard
        # ProcessReader. When we get data on this socket, we need to treat it
        # as a return code, or a failure.
        if not data.startswith('exited'):
            # Bad data, we want to signal that we are closing the connection
            # TODO: How?
            self.proc.childConnectionLost(self.name, "invalid data")
            self.close()
            # I don't know what to put here if we get bogus data, but I *do*
            # want to say that the process is now considered dead to me
            log.err('Got invalid exit information: %r' % (data,))
            exit_status = (255 << 8)
        else:
            exit_status = int(data.split('\n')[1])
        self.proc.processEnded(exit_status)


class ExecOnlySession(DoNothingSession):
    """Conch session that only allows executing commands."""

    def __init__(self, avatar, reactor, environment=None):
        super(ExecOnlySession, self).__init__(avatar)
        self.reactor = reactor
        self.environment = environment
        self._transport = None

    @classmethod
    def getAvatarAdapter(klass, environment=None):
        from twisted.internet import reactor
        return lambda avatar: klass(avatar, reactor, environment)

    def closed(self):
        """See ISession."""
        if self._transport is not None:
            # XXX: JonathanLange 2010-04-15: This is something of an
            # abstraction violation. Apart from this line and its twin, this
            # class knows nothing about Bazaar.
            notify(BazaarSSHClosed(self.avatar))
            try:
                self._transport.signalProcess('HUP')
            except (OSError, process.ProcessExitedAlready):
                pass
            self._transport.loseConnection()

    def eofReceived(self):
        """See ISession."""
        if self._transport is not None:
            self._transport.closeStdin()

    def execCommand(self, protocol, command):
        """Executes `command` using `protocol` as the ProcessProtocol.

        See ISession.

        :param protocol: a ProcessProtocol, usually SSHSessionProcessProtocol.
        :param command: A whitespace-separated command line. The first token is
        used as the name of the executable, the rest are used as arguments.
        """
        try:
            executable, arguments = self.getCommandToRun(command)
        except ForbiddenCommand, e:
            self.errorWithMessage(protocol, str(e) + '\r\n')
            return
        log.msg('Running: %r, %r' % (executable, arguments))
        if self._transport is not None:
            log.err(
                "ERROR: %r already running a command on transport %r"
                % (self, self._transport))
        # XXX: JonathanLange 2008-12-23: This is something of an abstraction
        # violation. Apart from this line and its twin, this class knows
        # nothing about Bazaar.
        notify(BazaarSSHStarted(self.avatar))
        self._transport = self._spawn(
            protocol, executable, arguments, env=self.environment)

    def _spawn(self, protocol, executable, arguments, env):
        return self.reactor.spawnProcess(
            protocol, executable, arguments, env=env)

    def getCommandToRun(self, command):
        """Return the command that will actually be run given `command`.

        :param command: A command line to run.
        :return: `(executable, arguments)` where `executable` is the name of an
            executable and arguments is a sequence of command-line arguments
            with the name of the executable as the first value.
        """
        args = command.split()
        return args[0], args


class RestrictedExecOnlySession(ExecOnlySession):
    """Conch session that only allows a single command to be executed."""

    def __init__(self, avatar, reactor, allowed_command,
                 executed_command_template, environment=None):
        """Construct a RestrictedExecOnlySession.

        :param avatar: See `ExecOnlySession`.
        :param reactor: See `ExecOnlySession`.
        :param allowed_command: The sole command that can be executed.
            This is the command as it comes from the client over the ssh
            protocol.
        :param executed_command_template: A Python format string for the actual
            command that will be run. '%(user_id)s' will be replaced with the
            'user_id' attribute of the current avatar.
        """
        ExecOnlySession.__init__(self, avatar, reactor, environment)
        self.allowed_command = allowed_command
        self.executed_command_template = executed_command_template

    @classmethod
    def getAvatarAdapter(klass, allowed_command, executed_command_template,
                         environment=None):
        from twisted.internet import reactor
        return lambda avatar: klass(
            avatar, reactor, allowed_command, executed_command_template,
            environment)

    def getCommandToRun(self, command):
        """As in ExecOnlySession, but only allow a particular command.

        :raise ForbiddenCommand: when `command` is not the allowed command.
        """
        if command != self.allowed_command:
            raise ForbiddenCommand("Not allowed to execute %r." % (command,))
        return ExecOnlySession.getCommandToRun(
            self, self.executed_command_template
            % {'username': self.avatar.username})


class SmartServerLauncher:
    """A class that is configured with the paths and names."""

    def __init__(self, bzr_command, hostname):
        """Create a launcher.

        :param bzr_command: The full command to invoke bzr on the machine.
        :param hostname: The hostname to use in the BZR_EMAIL.
        """
        self.bzr_command = bzr_command
        self.hostname = hostname

    def __call__(self, avatar):
        from twisted.internet import reactor

        # sc-serve comes from the sloecode plugin (yet todo).
        args = " sc-serve --inet %(username)s"
        command = self.bzr_command + args
        allowed_command = 'bzr serve --inet --directory=/ --allow-writes'

        environment = dict(os.environ)
        environment['BZR_EMAIL'] = '%s@%s' % (avatar.username, self.hostname)
        environment['BZR_PLUGIN_PATH'] = get_BZR_PLUGIN_PATH_for_subprocess()

        return RestrictedExecOnlySession(
            avatar, reactor, allowed_command, command, environment=environment)
